@@ -90,6 +90,31 @@ grecs_strdup(const char *str)
 	return strcpy(newstr, str);
 }
 
+char *
+grecs_strndup(const char *str, size_t len)
+{
+	char *newstr = strndup(str, len);
+	if (!newstr)
+		grecs_alloc_die();
+	return newstr;
+}
+
+void *
+grecs_2nrealloc(void *p, size_t *pn, size_t s)
+{
+	size_t n = *pn;
+
+	if (!p) {
+		if (n == 0)
+			n = 64 / s + 1;
+	} else if ((size_t) -1 / 3 * 2 / s <= n)
+		grecs_alloc_die();
+	else
+		n += (n + 1) / 2;
+	*pn = n;
+	return grecs_realloc(p, n * s);
+}
+
 void
 grecs_alloc_die(void)
 {
