@@ -473,9 +473,9 @@ json_object_set_string (struct json_value *obj,
 }
 
 void
-json_object_set_number (struct json_value *obj, char const *name, double val)
+json_object_set_number (struct json_value *obj, char const *name, long val)
 {
-  json_object_set (obj, name, json_new_number (val));
+  json_object_set (obj, name, json_new_long (val));
 }
 
 void
@@ -1817,7 +1817,8 @@ json_to_pcond (struct ctlio *io, struct json_value *val,
 	*pnode = pcond_node_alloc (val->b ? pcond_true : pcond_false);
 	break;
 	
-      case json_number:
+      case json_long:
+      case json_double:
       case json_string:
       case json_array:
 	ctlio_reply (io, 400, "bad conditional");
@@ -2449,13 +2450,13 @@ reload_diag_fun (grecs_locus_t const *locus, int err, int errcode,
       json_object_set (msg, "file", ar);
 
       ar = json_new_array ();
-      json_array_append (ar, json_new_number (locus->beg.line));
-      json_array_append (ar, json_new_number (locus->end.line));
+      json_array_append (ar, json_new_long (locus->beg.line));
+      json_array_append (ar, json_new_long (locus->end.line));
       json_object_set (msg, "line", ar);
 
       ar = json_new_array ();
-      json_array_append (ar, json_new_number (locus->beg.col));
-      json_array_append (ar, json_new_number (locus->end.col));
+      json_array_append (ar, json_new_long (locus->beg.col));
+      json_array_append (ar, json_new_long (locus->end.col));
       json_object_set (msg, "col", ar);
 
       json_object_set (msg, "error", json_new_bool (err));
@@ -2463,7 +2464,7 @@ reload_diag_fun (grecs_locus_t const *locus, int err, int errcode,
       if (errcode)
 	{
 	  json_object_set (msg, "syserrno",
-			   json_new_number (errno));
+			   json_new_long (errno));
 	  json_object_set (msg, "syserrstr",
 			   json_new_string (strerror (errno)));
 	}
