@@ -317,9 +317,257 @@ progman_waiting_p (void)
     }
   return 0;
 }
+
+enum deansify_state
+  {
+    deansi_off = -1,
+    deansi_init,
+    deansi_esc,
+    deansi_csi,
+    deansi_inter
+  };
+
+#define ESC 033
+
+static inline int
+is_param (int c)
+{
+  return 0x30 <= c && c <= 0x3f;
+}
+
+static inline int
+is_inter (int c)
+{
+  return 0x20 <= c && c <= 0x2f;
+}
+
+static inline int
+is_final (int c)
+{
+  return 0x40 <= c && c <= 0x7e;
+}
+
+struct deansify_ctx
+{
+  enum deansify_state state;
+  char hold[PIES_LOG_BUF_SIZE];
+  size_t len;
+};
+
+static inline void
+deansify_ctx_init (struct deansify_ctx *ctx, int enable)
+{
+  ctx->state = enable ? deansi_init : deansi_off;
+  ctx->len = 0;
+}
+
+static inline int
+deansify_ctx_add (struct deansify_ctx *ctx, int c)
+{
+  if (ctx->len == sizeof (ctx->hold))
+    return -1;
+  ctx->hold[ctx->len++] = c;
+  return 0;
+}
+
+static inline void
+deansify_ctx_clear (struct deansify_ctx *ctx)
+{
+  ctx->len = 0;
+}
+
+static inline void
+deansify_ctx_flush (struct deansify_ctx *ctx, char *text, size_t *idx)
+{
+  memcpy (text + *idx, ctx->hold, ctx->len);
+  *idx += ctx->len;
+  deansify_ctx_clear (ctx);
+}
+
+static inline void
+deansify_ctx_add_or_flush (struct deansify_ctx *ctx, int c,
+			   char *text, size_t *idx)
+{
+  if (deansify_ctx_add (ctx, c))
+    {
+      deansify_ctx_flush (ctx, text, idx);
+      ctx->state = deansi_init;
+      logmsg (LOG_INFO, "%s", _("ANSI sequence too long"));
+    }
+}
+
+size_t
+deansify (struct deansify_ctx *ctx, char *text, size_t size)
+{
+  size_t i, j;
+
+  for (i = j = 0; i < size && j < size; i++)
+    {
+      int c = text[i];
+      switch (ctx->state) {
+      case deansi_off:
+	return size;
+
+      case deansi_init:
+	if (c == ESC)
+	  {
+	    ctx->state = deansi_esc;
+	    deansify_ctx_add_or_flush (ctx, c, text, &j);
+	  }
+	else
+	  text[j++] = c;
+	break;
+
+      case deansi_esc:
+	if (c == '[')
+	  {
+	    ctx->state = deansi_csi;
+	    deansify_ctx_add_or_flush (ctx, c, text, &j);
+	  }
+	else
+	  {
+	    deansify_ctx_flush (ctx, text, &j);
+	    ctx->state = deansi_init;
+	  }
+	break;
+
+      case deansi_csi:
+	if (is_param (c))
+	  deansify_ctx_add_or_flush (ctx, c, text, &j);
+	else if (is_inter (c))
+	  {
+	    deansify_ctx_add_or_flush (ctx, c, text, &j);
+	    ctx->state = deansi_inter;
+	  }
+	else if (is_final (c))
+	  {
+	    deansify_ctx_clear (ctx);
+	    ctx->state = deansi_init;
+	  }
+	else
+	  {
+	    deansify_ctx_flush (ctx, text, &j);
+	    ctx->state = deansi_init;
+	  }
+	break;
+
+      case deansi_inter:
+	if (is_inter (c))
+	  deansify_ctx_add_or_flush (ctx, c, text, &j);
+	else if (is_final (c))
+	  {
+	    deansify_ctx_clear (ctx);
+	    ctx->state = deansi_init;
+	  }
+	else
+	  {
+	    deansify_ctx_flush (ctx, text, &j);
+	    ctx->state = deansi_init;
+	  }
+      }
+    }
+  return j;
+}
+
+struct read_buffer
+{
+  struct prog *master;
+  int stream;
+  char text[PIES_LOG_BUF_SIZE];
+  size_t len;
+  struct deansify_ctx dctx;
+  union
+  {
+    int overflow;  /* Overflow marker, for syslog. */
+    int fd;        /* Output file descriptor, for file. */
+  };
+};
+
+static struct read_buffer *
+read_buffer_new (struct prog *master, int stream, int deansify)
+{
+  struct read_buffer *rb = grecs_zalloc (sizeof (*rb));
+  rb->master = master;
+  rb->stream = stream;
+  rb->len = 0;
+  deansify_ctx_init (&rb->dctx, deansify);
+  return rb;
+}
+
+static void
+read_buffer_free (void *p)
+{
+  struct read_buffer *rb = p;
+  rb->master->v.p.redir[rb->stream] = -1;
+  free (rb);
+}
+
+static char *retr_stream_name[] = {
+  [RETR_OUT] = "stdout",
+  [RETR_ERR] = "stderr"
+};
+
+static int
+output_to_file (int fd, void *data)
+{
+  struct read_buffer *rb = data;
+  char const *tag = prog_tag (rb->master);
+  ssize_t n;
+
+  n = read (fd, rb->text + rb->len, sizeof (rb->text) - rb->len);
+  if (n == -1)
+    {
+      if (errno != EINTR)
+	{
+	  logmsg (LOG_INFO, _("%s: %s while reading %s"),
+		  tag, strerror (errno), retr_stream_name[rb->stream]);
+	  close (fd);
+	  deregister_socket (fd);
+	  return -1;
+	}
+    }
+  else if (n == 0)
+    {
+      debug (3, (_("%s: EOF on %s"),
+		 tag, retr_stream_name[rb->stream]));
+      close (fd);
+      deregister_socket (fd);
+    }
+  else
+    {
+      rb->len += deansify (&rb->dctx, rb->text + rb->len, n);
+      n = write (rb->fd, rb->text, rb->len);
+      if (n == -1)
+	{
+	  if (errno != EINTR)
+	    {
+	      logmsg (LOG_INFO, _("%s: %s while reading %s"),
+		      tag, strerror (errno), retr_stream_name[rb->stream]);
+	      close (fd);
+	      deregister_socket (fd);
+	      return -1;
+	    }
+	}
+      else if (n == 0)
+	{
+	  debug (3, (_("%s: EOF writing %s"),
+		     tag, retr_stream_name[rb->stream]));
+	  close (fd);
+	  deregister_socket (fd);
+	  return -1;
+	}
+      else
+	{
+	  if (n < rb->len)
+	    memmove (rb->text, rb->text + n, rb->len - n);
+	  rb->len -= n;
+	}
+    }
+  return 0;
+}
 
 int
-redirect_to_file (struct prog *master, int stream)
+redirect_to_file (struct prog *master, int stream, int deansify, int *rfd)
 {
   struct passwd *pw;
   int fd = open (master->v.p.comp->redir[stream].v.file, O_RDWR|O_CREAT,
@@ -349,20 +597,32 @@ redirect_to_file (struct prog *master, int stream)
 		master->v.p.comp->redir[stream].v.file,
 		strerror (errno));
     }
+
+  if (deansify)
+    {
+      int p[2];
+      struct read_buffer *rb;
+
+      if (pipe (p))
+	{
+	  logmsg (LOG_CRIT, "pipe: %s", strerror (errno));
+	  close (fd);
+	  return -1;
+	}
+
+      rb = read_buffer_new (master, stream, deansify);
+      rb->fd = fd;
+      register_socket (p[0], output_to_file, NULL, NULL, rb, read_buffer_free);
+      *rfd = p[0];
+      return p[1];
+    }
+  else
+    *rfd = -1;
   return fd;
 }
 
-struct read_buffer
-{
-  struct prog *master;
-  int stream;
-  char text[PIES_LOG_BUF_SIZE];
-  size_t len;
-  int overflow;
-};
-
 static int
-redirect_read (int fd, void *data)
+output_syslog (int fd, void *data)
 {
   struct read_buffer *rb = data;
   int n;
@@ -370,10 +630,6 @@ redirect_read (int fd, void *data)
   char const *tag = prog_tag (rb->master);
   pid_t pid = rb->master->pid;
   char *p;
-  static char *retr_stream_name[] = {
-    [RETR_OUT] = "stdout",
-    [RETR_ERR] = "stderr"
-  };
 
   n = read (fd, rb->text + rb->len, sizeof (rb->text) - rb->len);
   if (n == -1)
@@ -382,7 +638,6 @@ redirect_read (int fd, void *data)
 	{
 	  logmsg (LOG_INFO, _("%s: %s while reading %s"),
 		  tag, strerror (errno), retr_stream_name[rb->stream]);
-	  rb->master->v.p.redir[rb->stream] = -1;
 	  close (fd);
 	  deregister_socket (fd);
 	  return -1;
@@ -392,13 +647,12 @@ redirect_read (int fd, void *data)
     {
       debug (3, (_("%s: EOF on %s"),
 		 tag, retr_stream_name[rb->stream]));
-      rb->master->v.p.redir[rb->stream] = -1;
       close (fd);
       deregister_socket (fd);
     }
   else
     {
-      rb->len += n;
+      rb->len += deansify (&rb->dctx, rb->text + rb->len, n);
 
       while (rb->len)
 	{
@@ -429,31 +683,22 @@ redirect_read (int fd, void *data)
   return 0;
 }
 
-static void
-read_buffer_free (void *p)
-{
-  free (p);
-}
-
 int
-redirect_to_syslog (struct prog *master, int stream, int *fd)
+redirect_to_syslog (struct prog *master, int stream, int deansify, int *fd)
 {
   struct read_buffer *rb;
   int p[2];
-  
+
   if (pipe (p))
     {
       logmsg (LOG_CRIT, "pipe: %s", strerror (errno));
       return -1;
     }
 
-  rb = grecs_zalloc (sizeof (*rb));
-  rb->master = master;
-  rb->stream = stream;
-  rb->len = 0;
+  rb = read_buffer_new (master, stream, deansify);
   rb->overflow = 0;
-  
-  register_socket (p[0], redirect_read, NULL, NULL, rb, read_buffer_free);
+
+  register_socket (p[0], output_syslog, NULL, NULL, rb, read_buffer_free);
   *fd = p[0];
   return p[1];
 }
@@ -468,6 +713,7 @@ free_redirector (struct redirector *rp)
 int
 open_redirector (struct prog *master, int stream, int *fd)
 {
+  int deansify = master->v.p.comp->flags & CF_DEANSIFY;
   switch (master->v.p.comp->redir[stream].type)
     {
     case redir_null:
@@ -475,11 +721,10 @@ open_redirector (struct prog *master, int stream, int *fd)
       break;
       
     case redir_file:
-      *fd = -1;
-      return redirect_to_file (master, stream);
+      return redirect_to_file (master, stream, deansify, fd);
 
     case redir_syslog:
-      return redirect_to_syslog (master, stream, fd);
+      return redirect_to_syslog (master, stream, deansify, fd);
     }
   return -1;
 }
